/***************************************************************************
   par UTPECA d'après Rui Santos
   Complete project details at 
   https://RandomNerdTutorials.com/esp-now-one-to-many-esp32-esp8266/
   https://randomnerdtutorials.com/esp-now-esp32-arduino-ide/
   https://randomnerdtutorials.com/esp-now-esp8266-nodemcu-arduino-ide/  
   https://randomnerdtutorials.com/esp32-esp-now-wi-fi-web-server/ 
   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files.
  
   The above copyright notice and this permission notice shall be included in all
   copies or substantial portions of the Software.
  
   Croquis récepteur pour la mise sous tension des sections de voie :
     - fonctionne avec le croquis émetteur TCObyUTPECA4a1
****************************************************************************/

/*--------------------------------------------------------------------------
   Import required libraries
  --------------------------------------------------------------------------*/
#ifdef ESP32
  #include <esp_now.h>
  #include <esp_wifi.h>
  #include <WiFi.h>
#else
  #include <espnow.h>
  #include <ESP8266WiFi.h>
#endif
#define DEBUG

/*--------------------------------------------------------------------------
 * RECUPERATION CHANNEL DE L'EMETTEUR
---------------------------------------------------------------------------*/
constexpr char WIFI_SSID[] = "... le nom de votre routeur ...";
int32_t getWiFiChannel(const char *ssid)
{
  if (int32_t n = WiFi.scanNetworks())
  {
    for (uint8_t i = 0; i < n; i++)
    {
      if (!strcmp(ssid, WiFi.SSID(i).c_str()))
      {
        return WiFi.channel(i);
      }
    }
  }
  return 0;
}

/*--------------------------------------------------------------------------
   DONNEES PROTOCOLE ESP-NOW
  --------------------------------------------------------------------------*/
// interface de commande des appareils et des sections de voie
struct comm
{
  char   typRcp;       // type de récepteur (A = appareil, K = section)
  String nomKod;       // nom de l'appareil ou de la section

  union
  {
    struct             // Appareil de voie
    {
      byte    numApp;  // numéro appareil (de 1 à 8)
      char    typApp;  // type de l'appareil ('A', 'J', 'T', 'S')
      char    etaApp;  // état de l'appareil ( de '1' à '4')
      byte    noMot1;  // No du 1er moteur (de 1 à 8)
      byte    noMot2;  // No du 2me moteur (de 1 à 8)
    };
    struct             // Section de voie
    {
      char    etaSek;  // état : 0->rien, 1->alimenté
      byte    no_Rel;  // No du relais (de 1 à 8)
      byte    id_Rcp;  // No récepteur (rang+1 des macAdr)
      byte    dummy1;
      byte    dummy2;
    };
  };
};
//Create a struct_message called myComm
struct comm myComm;
struct comm oldComm;

//Numéro du récepteur (rang+1 des adresses Mac chez l'émetteur
byte    no_Rcp  = 3;

/*--------------------------------------------------------------------------
 * GESTION DU 74HC595 
---------------------------------------------------------------------------*/
#ifdef ESP32
const    byte verrou     = 2;     // 74HC595 : ST_CP (latch)
const    byte donnee     = 4;     // 74HC595 : DS (data)
const    byte horloge    = 0;     // 74HC595 : SH_CP (clock)                                 
#else
const    byte verrou     = 13;    // 74HC595 : ST_CP (latch) D7
const    byte donnee     = 14;    // 74HC595 : DS (data)     D5
const    byte horloge    = 12;    // 74HC595 : SH_CP (clock) D6
#endif

         uint8_t  u_Alim = 0b00000000;             // Data pour le 74HC595
const    uint8_t  z_Msk  = 0b10000000;             // Masque

/*********************************************************************/
void AlimenteSection() 
/*********************************************************************/
{
  String   nomKod  = myComm.nomKod;
  char     etaSek  = myComm.etaSek;
  byte     no_App  = myComm.numApp;
  int      no_Rel  = myComm.no_Rel - 1;            // No relais (de 1 à 8)
  uint8_t  z_Msk1  = z_Msk >> no_Rel;
  uint8_t  w_Alim;
  if (nomKod.indexOf("RAZ") < 0)
  { 
    w_Alim = u_Alim | z_Msk1;
    no_Rcp = myComm.id_Rcp; 
  }
  else
  { 
    w_Alim = 0b00000000;
    no_Rcp = 0; 
  }

#ifdef DEBUG 
  delay(100);
  Serial.print   ( myComm.nomKod);
  Serial.print   ( " ==> Av = ");
  Serial.print   ( u_Alim, BIN);
  Serial.print   ( "  Ap = " );
  Serial.println ( w_Alim, BIN);  
#endif

  // Envoi données relais au 74HC595
  //===================================================================
  u_Alim = w_Alim;
  digitalWrite(verrou, LOW);                  // mettre le verrou
  shiftOut(donnee, horloge, LSBFIRST, u_Alim);
  digitalWrite(verrou, HIGH);                 // oter le verrou
}

#ifdef ESP32
/*********************************************************************/
void OnDataRecv(const uint8_t * mac, const uint8_t *incomingData, int len) 
/*********************************************************************/
//callback function that will be executed when data is received
{
  memcpy(&myComm, incomingData, sizeof(myComm));
  Serial.print   ( myComm.nomKod);
  Serial.print   ( " ==> ");
  Serial.print   ( myComm.etaSek == '0' ? "hors" : "sous");
  Serial.println ( " tension");
}
#else
/*********************************************************************/
void OnDataRecv(uint8_t * mac, uint8_t *incomingData, uint8_t len) 
/*********************************************************************/
{
  memcpy(&myComm, incomingData, sizeof(myComm));
  Serial.print   ( myComm.nomKod);
  Serial.print   ( " ==> ");
  Serial.print   ( myComm.etaSek == '0' ? "hors" : "sous");
  Serial.println ( " tension");
}
#endif 
 
/*********************************************************************/
void setup() 
/*********************************************************************/
{
  //Initialize Serial Monitor for debugging purposes
  //------------------------------------------------------------------
  Serial.begin(115200);
#ifdef ESP32
  Serial.print("### TCO-Web/ReceptSec (v4.50) sur ");
  Serial.println("ESP32 ###");
#else
  Serial.println("");
  Serial.print("### TCO-Web/ReceptSec (v4.50) sur ");
  Serial.println("ESP8266 ###");
#endif

  // Préparation du 74HC595 
  //------------------------------------------------------------------
  pinMode(horloge, OUTPUT);
  pinMode(verrou,  OUTPUT);
  pinMode(donnee,  OUTPUT);
  
  myComm.nomKod  = "RAZ_";
  myComm.etaSek  = '0';
  oldComm.nomKod = "RAZ_";
  oldComm.etaSek = '0';

  // Connection à ESP-NOW
  //------------------------------------------------------------------
  WiFi.mode(WIFI_STA);
  int32_t channel = getWiFiChannel(WIFI_SSID);
  Serial.print   ( "WiFi SSID =  " );
  Serial.println ( WIFI_SSID );
  Serial.print   ( "Old WiFi channel : " );
  Serial.println ( WiFi.channel() );
  Serial.print   ( "New WiFi channel : " );
  Serial.println ( channel );
  
#ifdef ESP32
  esp_wifi_set_promiscuous(true);
  esp_wifi_set_channel(channel, WIFI_SECOND_CHAN_NONE);
  esp_wifi_set_promiscuous(false);
#else
  wifi_promiscuous_enable(1);
  wifi_set_channel(channel);
  wifi_promiscuous_enable(0);
#endif


//WiFi.disconnect();
//Serial.println("");
  Serial.print("Initialisation ESP-NOW : ");
  
#ifdef ESP32
  if (esp_now_init() != ESP_OK) {
    Serial.println("Erreur");
    return;
  }
  Serial.println("Ok");
  // fonction callback
  esp_now_register_recv_cb(OnDataRecv);
#else
  if (esp_now_init() != 0) {
    Serial.println("Erreur");
    return;
  }
  Serial.println("Ok");
  // fonction callback
  esp_now_set_self_role(ESP_NOW_ROLE_CONTROLLER);
  esp_now_register_recv_cb(OnDataRecv);
#endif

  Serial.print ( "ReceptSec sur  Mac[" );
  Serial.print ( WiFi.macAddress() );
  Serial.println("] prêt à recevoir.");
}
 
/*********************************************************************/
void loop() 
/*********************************************************************/
{
  // test : entrée d'une commande de mise sous tension
  if (myComm.nomKod != oldComm.nomKod ||
      myComm.etaSek != oldComm.etaSek)
  {
     AlimenteSection();
     
     oldComm.nomKod = myComm.nomKod;
     oldComm.etaSek = myComm.etaSek;
     delay(100);
  }
}
